import itertools
import logging
import typing as t

import numpy as np
import pandas as pd
import plotly.express as px
from pydantic import ConfigDict, Field

from zndraw.base import Extension, MethodsCollection

try:
    from zndraw.analyse import mda  # noqa: F401
except ImportError:
    # mdanalysis is not installed
    pass


log = logging.getLogger(__name__)


def _schema_from_atoms(schema, cls):
    return cls.model_json_schema_from_atoms(schema)


class DihedralAngle(Extension):
    def run(self, vis):
        atoms_lst = list(vis)
        dihedral_angles = []

        if len(vis.selection) != 4:
            raise ValueError("Please select exactly 4 atoms")
        for atoms in atoms_lst:
            dihedral_angles.append(
                atoms.get_dihedrals(indices=[vis.selection], mic=True)[0]
            )
        df = pd.DataFrame(
            {"step": list(range(len(atoms_lst))), "dihedral": dihedral_angles}
        )
        fig = px.line(df, x="step", y="dihedral", render_mode="svg")
        vis.figure = fig.to_json()


class Distance(Extension):
    smooth: bool = False

    def run(self, vis):
        atoms_lst, ids = list(vis), vis.selection
        distances = {}
        for x in itertools.combinations(ids, 2):
            distances[f"{tuple(x)}"] = []
        for atoms in atoms_lst:
            positions = atoms.get_positions()
            for x in itertools.combinations(ids, 2):
                distances[f"{tuple(x)}"].append(
                    np.linalg.norm(positions[x[0]] - positions[x[1]])
                )

        df = pd.DataFrame({"step": list(range(len(atoms_lst)))} | distances)

        fig = px.line(
            df,
            x="step",
            y=df.columns,
            title="Distance between selected particles",
            render_mode="svg",  # This is important, otherwise openGL will be used
            # and there can/will be issues with three.js
        )
        if self.smooth:
            smooth_df = df.rolling(window=100).mean().dropna()
            for col in smooth_df.columns:
                if col != "step":
                    fig.add_scatter(
                        x=smooth_df["step"], y=smooth_df[col], name=f"smooth_{col}"
                    )
        vis.figure = fig.to_json()


class Properties2D(Extension):
    x_data: str
    y_data: str
    color: str
    fix_aspect_ratio: bool = True

    model_config = ConfigDict(json_schema_extra=_schema_from_atoms)

    @classmethod
    def model_json_schema_from_atoms(cls, schema: dict) -> dict:
        ATOMS = cls.get_atoms()
        log.debug(f"GATHERING PROPERTIES FROM {ATOMS=}")
        try:
            available_properties = list(ATOMS.calc.results)
            available_properties += list(ATOMS.arrays)
            available_properties += ["step"]
            schema["properties"]["x_data"]["enum"] = available_properties
            schema["properties"]["y_data"]["enum"] = available_properties
            schema["properties"]["color"]["enum"] = available_properties
        except AttributeError:
            pass
        return schema

    def run(self, vis):
        atoms_lst = list(vis)
        log.info(f"run {self}")

        if self.x_data == "step":
            x_data = list(range(len(atoms_lst)))
        else:
            try:
                x_data = [x.calc.results[self.x_data] for x in atoms_lst]
            except KeyError:
                x_data = [x.arrays[self.x_data] for x in atoms_lst]

        if self.y_data == "step":
            y_data = list(range(len(atoms_lst)))
        else:
            try:
                y_data = [x.calc.results[self.y_data] for x in atoms_lst]
            except KeyError:
                y_data = [x.arrays[self.y_data] for x in atoms_lst]

        if self.color == "step":
            color = list(range(len(atoms_lst)))
        else:
            try:
                color = [x.calc.results[self.color] for x in atoms_lst]
            except KeyError:
                color = [x.arrays[self.color] for x in atoms_lst]

        y_data = np.array(y_data).reshape(-1)
        x_data = np.array(x_data).reshape(-1)
        color = np.array(color).reshape(-1)

        df = pd.DataFrame({self.x_data: x_data, self.y_data: y_data, self.color: color})
        fig = px.scatter(
            df, x=self.x_data, y=self.y_data, color=self.color, render_mode="svg"
        )
        if self.fix_aspect_ratio:
            fig.update_yaxes(
                scaleanchor="x",
                scaleratio=1,
            )
        vis.figure = fig.to_json()


class Properties1D(Extension):
    value: str
    smooth: bool = False

    model_config = ConfigDict(json_schema_extra=_schema_from_atoms)
    aggregation: t.Literal["mean", "median", "max", ""] = Field(
        "",
        description="For multidimensional data, aggregate over all dimensions, except the first one.",
    )

    @classmethod
    def model_json_schema_from_atoms(cls, schema: dict) -> dict:
        ATOMS = cls.get_atoms()
        try:
            available_properties = list(ATOMS.calc.results.keys())  # global ATOMS object
            log.debug(f"AVAILABLE PROPERTIES: {available_properties=}")
            schema["properties"]["value"]["enum"] = available_properties
        except AttributeError:
            print(f"{ATOMS=}")
        return schema

    def run(self, vis):
        vis.log("Downloading data...")
        atoms_lst = list(vis)
        data = np.array([x.calc.results[self.value] for x in atoms_lst])

        if data.ndim > 1:
            axis = tuple(range(1, data.ndim))
            if self.aggregation == "mean":
                data = np.mean(data, axis=axis)
            elif self.aggregation == "median":
                data = np.median(data, axis=axis)
            elif self.aggregation == "max":
                data = np.max(data, axis=axis)

        df = pd.DataFrame({"step": list(range(len(atoms_lst))), self.value: data})

        fig = px.line(df, x="step", y=self.value, render_mode="svg")

        if self.smooth:
            smooth_df = df.rolling(window=100).mean().dropna()
            for col in smooth_df.columns:
                if col != "step":
                    fig.add_scatter(
                        x=smooth_df["step"], y=smooth_df[col], name=f"smooth_{col}"
                    )

        vis.figure = fig.to_json()


methods = t.Union[
    Properties1D,
    DihedralAngle,
    Distance,
    Properties2D,
]


class Analysis(MethodsCollection):
    method: methods = Field(description="Analysis method", discriminator="discriminator")



import dataclasses
import logging
import typing as t

import ase
import numpy as np
import socketio
import znframe

from zndraw.base import Extension, ZnDrawBase
from zndraw.draw import Geometry, Object3D

log = logging.getLogger(__name__)


@dataclasses.dataclass
class ZnDraw(ZnDrawBase):
    url: str
    token: str | None = None
    auth_token: str | None = None

    socket: socketio.Client = dataclasses.field(
        default_factory=socketio.Client, repr=False
    )

    _modifiers: dict = dataclasses.field(default_factory=dict)
    _available: bool = True

    def __post_init__(self):
        def on_wakeup():
            if self._available:
                self.socket.emit("modifier:available", list(self._modifiers))

        self.url = self.url.replace("http", "ws")
        self.socket.on("connect", self._on_connect)
        self.socket.connect(self.url, wait_timeout=1)
        self.socket.on("modifier:run", self._run_modifier)
        self.socket.on("modifier:wakeup", on_wakeup)
        self.socket.on("room:log", lambda x: print(x))

    def _on_connect(self):
        self.socket.emit(
            "join",
            {
                "token": str(self.token),
                "auth_token": self.auth_token,
            },
        )

    def __getitem__(self, index) -> ase.Atoms | list[ase.Atoms]:
        single_item = isinstance(index, int)
        if single_item:
            index = [index]

        if isinstance(index, slice):
            index = list(range(*index.indices(len(self))))

        if any(x < 0 for x in index):
            raise IndexError("Index must be positive")
        if any(x >= len(self) for x in index):
            raise IndexError("Index out of range")

        data: dict = self.socket.call("room:frames:get", index)
        structures = [znframe.Frame(**x).to_atoms() for x in data.values()]
        return structures[0] if single_item else structures

    def __setitem__(self, index: int | list[int], value: ase.Atoms | list[ase.Atoms]):
        if isinstance(index, int):
            data = {index: znframe.Frame.from_atoms(value).to_json()}
        else:
            data = {
                i: znframe.Frame.from_atoms(val).to_json() for i, val in zip(index, value)
            }

        self.socket.emit("room:frames:set", data)

    def __len__(self) -> int:
        return int(self.socket.call("room:length:get"))

    def __delitem__(self, index: int | slice | list[int]):
        if isinstance(index, int):
            index = [index]
        if isinstance(index, slice):
            index = list(range(*index.indices(len(self))))
        self.socket.emit("room:frames:delete", index)

    def insert(self, index: int, value: ase.Atoms):
        self.socket.emit(
            "room:frames:insert",
            {"index": index, "value": znframe.Frame.from_atoms(value).to_json()},
        )

    @property
    def selection(self) -> list[int]:
        return self.socket.call("room:selection:get")["0"]
        data: dict = self.socket.call("room:selection:get")
        return data["0"]

    @selection.setter
    def selection(self, value: list[int]):
        self.socket.emit("room:selection:set", {"0": value})

    @property
    def step(self) -> int:
        return int(self.socket.call("room:step:get"))

    def log(self, message: str) -> None:
        self.socket.emit("room:log", message)

    @step.setter
    def step(self, value: int):
        if value < 0:
            raise ValueError("Step must be positive")
        if value >= len(self):
            raise ValueError("Step out of range")
        # Have only one step per room!
        # shared rooms are rare anyhow and making per-client steps
        # and room hosts is anoying
        # what about the camera?
        # or collect the steps of all clients in a dict
        # and save the host and go from there, also fine and not too much worker.
        self.socket.emit("room:step:set", value)

    @property
    def figure(self) -> str:
        return self.socket.call("analysis:figure:get")

    @figure.setter
    def figure(self, fig: str) -> None:
        self.socket.emit("analysis:figure:set", fig)

    @property
    def atoms(self) -> ase.Atoms:
        return self[self.step]

    @atoms.setter
    def atoms(self, atoms: ase.Atoms) -> None:
        self[[self.step]] = [atoms]

    @property
    def points(self) -> list[list[float]]:
        return np.array(self.socket.call("room:points:get")["0"])

    @points.setter
    def points(self, points: np.ndarray) -> None:
        self.socket.emit("room:points:set", {"0": np.array(points).tolist()})

    @property
    def bookmarks(self) -> dict:
        return {int(k): v for k, v in self.socket.call("room:bookmarks:get").items()}

    @bookmarks.setter
    def bookmarks(self, value: dict):
        self.socket.emit("room:bookmarks:set", value)

    @property
    def camera(self) -> dict:
        return self.socket.call("room:camera:get")

    @camera.setter
    def camera(self, value: dict):
        raise NotImplementedError("Changing camera position is currently not possible")
        self.socket.emit("room:camera:set", value)

    @property
    def geometries(self) -> list[Object3D]:
        # return self.socket.call("room:geometry:get")

        return [Geometry(method=x).method for x in self.socket.call("room:geometry:get")]

    @geometries.setter
    def geometries(self, value: list[Object3D]):
        self.socket.emit("room:geometry:set", [x.model_dump() for x in value])

    def register_modifier(
        self,
        cls: t.Type[Extension],
        run_kwargs: dict | None = None,
        public: bool = False,
        timeout: float = 60,
        use_frozen: bool = True,
    ):
        """Register a modifier class.

        Attributes
        ----------
        cls : UpdateScene
            The modifier class to register.
        run_kwargs : dict, optional
            Keyword arguments to pass to the run method of the modifier class.
        public : bool, optional
            Whether to enable the modifier for ALL sessions of the ZnDraw client,
            or just the session for the given token.
        timeout : float, optional
            Timeout for the modifier to run in seconds. The Webclient
            will alert the user if the modifier takes longer than this time and
            release the modify button (no further changes are expected, but they
            can happen).
        use_frozen : bool, default=True
            Whether to use the ZnDrawFrozen class to run the modifier.
            The frozen class only allows provides cached data and
            e.g. access to other steps than the current one is not possible.
            If set to false, a full ZnDraw instance will be created for the modifier.
            This can have a performance impact and may lead to timeouts.

        """
        if timeout < 1:
            raise ValueError("Timeout must be at least 1 second")
        if timeout > 300:
            log.critical(
                "Timeout is set to more than 300 seconds. Modifiers might be killed automatically."
            )
        if run_kwargs is None:
            run_kwargs = {}
        run_kwargs["timeout"] = timeout
        if cls.__name__ in self._modifiers:
            raise ValueError(f"Modifier {cls.__name__} already registered")

        self.socket.emit(
            "modifier:register",
            {
                "schema": cls.model_json_schema(),
                "name": cls.__name__,
                "public": public,
                "timeout": timeout,
            },
        )
        self._modifiers[cls.__name__] = {
            "cls": cls,
            "run_kwargs": run_kwargs,
            "public": public,
            "frozen": use_frozen,
        }
        if self._available:
            self.socket.emit("modifier:available", list(self._modifiers))

    def _run_modifier(self, data: dict):
        self._available = False
        room = data.pop("ZNDRAW_CLIENT_ROOM")
        vis = type(self)(url=self.url, token=room)

        try:
            # TODO: for public modifiers the vis object must not be in the same room, create a new one!!!!!
            vis.socket.emit("room:modifier:queue", 0)
            name = data["method"]["discriminator"]

            instance = self._modifiers[name]["cls"](**data["method"])
            instance.run(vis, **self._modifiers[name]["run_kwargs"])

            vis.socket.emit("room:modifier:queue", -1)
        finally:
            self._available = True
            self.socket.emit("modifier:available", list(self._modifiers))



import random
import typing as t

import networkx as nx
import numpy as np
from pydantic import Field

from zndraw import Extension, ZnDraw
from zndraw.base import MethodsCollection

try:
    from zndraw.select import mda  # noqa: F401
except ImportError:
    # mdanalysis is not installed
    pass


class SelectionBase(Extension):
    def run(self, vis: "ZnDraw") -> None:
        raise NotImplementedError()


class NoneSelection(SelectionBase):
    def run(self, vis) -> None:
        vis.selection = []


class All(SelectionBase):
    """Select all atoms."""

    def run(self, vis) -> None:
        atoms = vis[vis.step]
        vis.selection = list(range(len(atoms)))


class Invert(SelectionBase):
    def run(self, vis) -> None:
        atoms = vis[vis.step]
        selected_ids = vis.selection
        vis.selection = list(set(range(len(atoms))) - set(selected_ids))


class Range(SelectionBase):
    start: int = Field(0, description="Start index")
    end: int = Field(5, description="End index")
    step: int = Field(1, description="Step size")

    def run(self, vis) -> None:
        vis.selection = list(range(self.start, self.end, self.step))


class Random(SelectionBase):
    count: int = Field(..., description="Number of atoms to select")

    def run(self, vis) -> None:
        atoms = vis[vis.step]
        vis.selection = random.sample(range(len(atoms)), self.count)


class IdenticalSpecies(SelectionBase):
    def run(self, vis) -> None:
        atoms = vis[vis.step]
        selected_ids = vis.selection
        selected_ids = set(selected_ids)
        for idx in tuple(selected_ids):
            selected_symbol = atoms[idx].symbol
            selected_ids.update(
                idx for idx, atom in enumerate(atoms) if atom.symbol == selected_symbol
            )
        vis.selection = list(selected_ids)


class ConnectedParticles(SelectionBase):
    def run(self, vis) -> None:
        atoms = vis.atoms
        selected_ids = vis.selection
        total_ids = []
        try:
            edges = atoms.connectivity
            graph = nx.Graph()
            for edge in edges:
                node_a, node_b, weight = edge
                graph.add_edge(node_a, node_b, weight=weight)
        except AttributeError:
            return selected_ids

        for node_id in selected_ids:
            total_ids += list(nx.node_connected_component(graph, node_id))
            total_ids = np.array(total_ids)

        vis.selection = [x.item() for x in set(total_ids)]


class Neighbour(SelectionBase):
    """Select the nth order neighbours of the selected atoms."""

    order: int = Field(1, description="Order of neighbour")

    def run(self, vis) -> None:
        total_ids = []
        atoms = vis[vis.step]
        selected_ids = vis.selection
        try:
            graph = atoms.connectivity
        except AttributeError:
            return selected_ids

        for node_id in selected_ids:
            total_ids += list(
                nx.single_source_shortest_path_length(graph, node_id, self.order).keys()
            )

        vis.selection = list(set(total_ids))


class UpdateSelection(SelectionBase):
    """Reload Selection."""

    def run(self, vis: ZnDraw) -> None:
        vis.selection = vis.selection


# Order here matters
methods = t.Union[
    ConnectedParticles,
    NoneSelection,
    All,
    Invert,
    Range,
    Random,
    IdenticalSpecies,
    Neighbour,
]


class Selection(MethodsCollection):
    method: methods = Field(
        description="Selection method",
        discriminator="discriminator",
    )



import json
import logging
import uuid

import znsocket
from flask import current_app, request, session
from flask_socketio import emit, join_room
from redis import Redis

from zndraw.analyse import Analysis
from zndraw.draw import Geometry
from zndraw.modify import Modifier
from zndraw.scene import Scene
from zndraw.select import Selection
from zndraw.utils import get_cls_from_json_schema

from ..app import socketio as io
from ..tasks import run_analysis, run_geometry, run_modifier, run_selection

log = logging.getLogger(__name__)


@io.on("connect")
def connect():
    pass


@io.on("disconnect")
def disconnect():
    room = str(session["token"])
    r = current_app.config["redis"]

    if "name" in session:
        log.critical(f"disconnecting (webclient) {request.sid} from room {room}")
        r.srem(f"room:{room}:webclients", session["name"])
        emit("room:users:refresh", list(r.smembers(f"room:{room}:webclients")), to=room)
    else:
        log.critical(f"disconnecting (pyclient) {request.sid}")

        # Get all modifier names associated with the client from Redis
        room_modifier_names = r.smembers(f"room:{room}:modifiers")
        default_modifier_names = r.smembers("room:default:modifiers")

        for modifier in room_modifier_names:
            r.srem(f"room:{room}:modifiers:{modifier}", request.sid)
            if r.scard(f"room:{room}:modifiers:{modifier}") == 0:
                r.hdel(f"room:{room}:modifiers", modifier)
        for modifier in default_modifier_names:
            r.srem(f"room:default:modifiers:{modifier}", request.sid)
            if r.scard(f"room:default:modifiers:{modifier}") == 0:
                r.hdel("room:default:modifiers", modifier)

        # Emit an event to refresh the modifier schema
        emit("modifier:schema:refresh", broadcast=True)


@io.on("webclient:connect")
def webclient_connect():
    try:
        token = session["token"]
    except KeyError:
        token = uuid.uuid4().hex[:8] if current_app.config["USE_TOKEN"] else "main"
        session["token"] = token

    room = str(session["token"])
    join_room(room)  # rename token to room or room_token

    session["name"] = uuid.uuid4().hex[:8]

    r = current_app.config["redis"]
    r.sadd(f"room:{room}:webclients", session["name"])

    emit("room:users:refresh", list(r.smembers(f"room:{room}:webclients")), to=room)
    # set step, camera, bookmarks, points

    log.critical(f"connecting (webclient) {request.sid} to {room}")

    emit("selection:schema", Selection.updated_schema())
    emit("modifier:schema:refresh", to=request.sid)
    emit("scene:schema", Scene.updated_schema())
    emit("geometry:schema", Geometry.updated_schema())
    emit("analysis:schema:refresh", to=request.sid)

    return {"name": session["name"], "room": room}


@io.on("join")  # rename pyclient:connect
def join(data: dict):
    """
    Arguments:
        data: {"token": str, "auth_token": str}
    """
    # TODO: prohibt "token" to be "default"

    if current_app.config["AUTH_TOKEN"] is None:
        session["authenticated"] = True
    else:
        session["authenticated"] = data["auth_token"] == current_app.config["AUTH_TOKEN"]
    token = data["token"]
    session["token"] = token
    room = str(session["token"])

    join_room(room)
    log.critical(f"connecting (pyclient) {request.sid} to {room}")
    # join_room(f"pyclients_{token}")


@io.on("room:frames:get")
def room_frames_get(frames: list[int]) -> dict[int, dict]:
    # print(f"requesting frames: {frames}")
    if len(frames) == 0:
        return {}
    r: Redis = current_app.config["redis"]
    room = session.get("token")

    if r.exists(f"room:{room}:frames"):
        data = znsocket.List(r, f"room:{room}:frames")[frames]

    else:
        try:
            data = znsocket.List(r, "room:default:frames")[frames]
        except IndexError:
            data = []

    return {idx: json.loads(d) for idx, d in zip(frames, data) if d is not None}


@io.on("room:frames:set")
def room_frames_set(data: dict[int, str]):
    r: Redis = current_app.config["redis"]
    room = session.get("token")

    # add = {}
    # remove = []
    lst = znsocket.List(r, f"room:{room}:frames")

    if not r.exists(f"room:{room}:frames"):
        default_lst = znsocket.List(r, "room:default:frames")
        # TODO: using a redis copy action would be faster
        lst.extend(default_lst)

    lst[list(data)] = [d for d in data.values()]

    emit("room:frames:refresh", [int(x) for x in data], to=room)


@io.on("room:all:frames:refresh")
def room_all_frames_refresh(indices: list[int]):
    emit("room:frames:refresh", [int(x) for x in indices], broadcast=True)


@io.on("room:frames:delete")
def room_frames_delete(frames: list[int]):
    r: Redis = current_app.config["redis"]
    room = session.get("token")
    lst = znsocket.List(r, f"room:{room}:frames")
    if not r.exists(f"room:{room}:frames"):
        default_lst = znsocket.List(r, "room:default:frames")
        # TODO: using a redis copy action would be faster
        lst.extend(default_lst)
    del lst[frames]
    # TODO how to update here?
    emit("room:frames:refresh", [int(x) for x in frames], to=room)


@io.on("room:frames:insert")
def room_frames_insert(data: dict):
    index = data.pop("index")
    value = data.pop("value")
    r: Redis = current_app.config["redis"]
    room = session.get("token")

    lst = znsocket.List(r, f"room:{room}:frames")
    if not r.exists(f"room:{room}:frames"):
        default_lst = znsocket.List(r, "room:default:frames")
        # TODO: using a redis copy action would be faster
        lst.extend(default_lst)
    lst.insert(index, value)

    # not sure how to update, insert requires everything to be updated after the insertion
    # can be done custom on the client side to avoid resending everything
    emit("room:frames:refresh", [int(x) for x in data], to=room)


@io.on("room:length:get")
def room_frames_length_get() -> int:
    room = session.get("token")
    r: Redis = current_app.config["redis"]
    room_key = (
        f"room:{room}:frames"
        if r.exists(f"room:{room}:frames")
        else "room:default:frames"
    )
    return len(znsocket.List(r, room_key))


@io.on("modifier:register")
def modifier_register(data: dict):
    """Register the modifier."""
    if data["public"] and not session["authenticated"]:
        log.critical("Unauthenticated user tried to register a default modifier.")
        return

    r: Redis = current_app.config["redis"]
    room = session.get("token")

    data["ZNDRAW_CLIENT_SID"] = request.sid

    if data.pop("public"):
        r.hset("room:default:modifiers", data["name"], json.dumps(data))
        r.sadd(f"room:default:modifiers:{data['name']}", request.sid)
    else:
        r.hset(f"room:{room}:modifiers", data["name"], json.dumps(data))
        r.sadd(f"room:{room}:modifiers:{data['name']}", request.sid)

    # only if public this is required
    emit("modifier:schema:refresh", broadcast=True)


@io.on("modifier:schema")
def modifier_schema():
    r: Redis = current_app.config["redis"]
    room = session.get("token")

    modifiers: dict = r.hgetall(f"room:{room}:modifiers")
    modifiers |= r.hgetall("room:default:modifiers")
    # reconstruct them to get the schema
    classes = []
    for modifier in modifiers.values():
        modifier = json.loads(modifier)
        cls = get_cls_from_json_schema(modifier["schema"], modifier["name"])
        classes.append(cls)

    emit("modifier:schema", Modifier.updated_schema(extensions=classes), to=request.sid)


@io.on("draw:schema")
def draw_schema():
    return Geometry.updated_schema()


@io.on("scene:schema")
def scene_schema():
    emit("scene:schema", Scene.updated_schema(), to=request.sid)


@io.on("selection:schema")
def selection_schema():
    emit("selection:schema", Selection.updated_schema(), to=request.sid)


@io.on("analysis:schema")
def analysis_schema():
    emit("analysis:schema", Analysis.updated_schema(), to=request.sid)


@io.on("modifier:run")
def modifier_run(data: dict):
    room = session.get("token")
    emit("room:modifier:queue", 1, to=room)  # TODO: find the correct queue position
    url = f"http://127.0.0.1:{current_app.config['PORT']}"

    r: Redis = current_app.config["redis"]

    name = data["method"]["discriminator"]

    public = r.smembers(f"room:default:modifiers:{name}")
    privat = r.smembers(f"room:{room}:modifiers:{name}")

    data["ZNDRAW_CLIENT_ROOM"] = room

    if len(public):
        # The modifier was registered with public=True
        r.rpush(f"modifier:queue:{name}", json.dumps(data))
    elif len(privat):
        # The modifier was registered with public=False
        r.rpush(f"modifier:queue:{room}:{name}", json.dumps(data))
    else:
        # This would be the queue for default modifiers.
        # but they are queued using celery directly.
        # so no need for redis queue.
        pass

    clients: set[str] = public | privat
    if len(clients):
        for sid in clients:
            # kindly ask every client if they are available
            emit("modifier:wakeup", to=sid)
    else:
        run_modifier.delay(url, room, data)


@io.on("room:modifier:queue")
def room_modifier_run(data: int):
    """Forwarding finished message."""
    room = session.get("token")
    emit("room:modifier:queue", data, to=room)


@io.on("modifier:available")
def modifier_available(modifier_names: list[str]) -> None:
    """Update state of registered modifier classes."""
    r: Redis = current_app.config["redis"]
    room = session.get("token")  # TODO: Why use get, token should always be set

    for name in modifier_names:
        public = r.smembers(f"room:default:modifiers:{name}")
        privat = r.smembers(f"room:{room}:modifiers:{name}")

        if request.sid in public:
            # TODO: !!! there will be an issue if you register a privat modifier "a"
            # and then someone else registers a public modifier "a"
            public_task = r.lpop(f"modifier:queue:{name}")
            if public_task:
                log.debug(f"running public task {public_task} on {request.sid}")
                emit("modifier:run", json.loads(public_task), to=request.sid)
                return

        if request.sid in privat:
            privat_task = r.lpop(f"modifier:queue:{room}:{name}")
            if privat_task:
                log.debug(f"running private task {privat_task} on {request.sid}")
                emit("modifier:run", json.loads(privat_task), to=request.sid)
                return

    log.debug(f"No task available for {modifier_names}")


@io.on("analysis:run")
def analysis_run(data: dict):
    room = session.get("token")
    emit("room:analysis:queue", 1, to=room)  # TODO: find the correct queue position
    url = f"http://127.0.0.1:{current_app.config['PORT']}"
    run_analysis.delay(url, room, data)


@io.on("room:analysis:queue")
def room_analysis_run(data: int):
    """Forwarding finished message."""
    room = session.get("token")
    emit("room:analysis:queue", data, to=room)


@io.on("room:geometry:queue")
def room_geometry_run(data: int):
    """Forwarding finished message."""
    room = session.get("token")
    emit("room:geometry:queue", data, to=room)


@io.on("geometry:run")
def geometry_run(data: dict):
    room = session.get("token")
    emit("geometry:run:enqueue", to=room)
    url = f"http://127.0.0.1:{current_app.config['PORT']}"
    run_geometry.delay(url, room, data)


@io.on("room:geometry:set")
def room_geometry_set(data: list):
    r: Redis = current_app.config["redis"]
    room = session.get("token")

    # # add = {}
    # # remove = []
    lst = znsocket.List(r, f"room:{room}:geometries")
    del lst[:]
    lst.extend(data)
    emit("room:geometry:set", data, to=room, include_self=False)


@io.on("room:geometry:get")
def room_geometry_get():
    r: Redis = current_app.config["redis"]
    room = session.get("token")

    return list(znsocket.List(r, f"room:{room}:geometries"))


@io.on("analysis:figure:set")
def analysis_figure_set(data: dict):
    # This is currently using push and the figure is not stored
    room = session.get("token")
    r: Redis = current_app.config["redis"]
    r.set(f"room:{room}:analysis:figure", json.dumps(data))
    emit(
        "analysis:figure:set",
        data,
        to=room,
    )


@io.on("analysis:figure:get")
def analysis_figure_get() -> dict:
    room = session.get("token")
    r: Redis = current_app.config["redis"]
    return json.loads(r.get(f"room:{room}:analysis:figure"))


@io.on("selection:run")
def selection_run(data: dict):
    room = session.get("token")
    emit("room:selection:queue", 1, to=room)  # TODO: find the correct queue position
    url = f"http://127.0.0.1:{current_app.config['PORT']}"
    run_selection.delay(url, room, data)


# @io.on("selection:run:finished")
# def selection_run_finished():
#     """Forwarding finished message."""
#     room = session.get("token")
#     emit("selection:run:finished", to=room)


# @io.on("selection:run:running")
# def selection_run_running():
#     """Forwarding running method."""
#     room = session.get("token")
#     emit("selection:run:running", to=room)


@io.on("room:selection:queue")
def room_selection_run(data: int):
    """Forwarding finished message."""
    room = session.get("token")
    emit("room:selection:queue", data, to=room)


@io.on("room:alert")
def room_alert(msg: str):
    """Forward the alert message to every client in the room"""
    # TODO: identify the source client.
    room = session.get("token")
    emit("room:alert", msg, to=room)


@io.on("room:log")
def room_log(msg: str):
    """Forward the alert message to every client in the room"""
    room = session.get("token")
    emit("room:log", msg, to=room)


@io.on("room:selection:set")
def room_selection_set(data: dict[str, list[int]]):
    r: Redis = current_app.config["redis"]
    room = session.get("token")
    r.hmset(f"room:{room}:selection", {k: json.dumps(v) for k, v in data.items()})
    emit("room:selection:set", data, to=room, include_self=False)


@io.on("room:selection:get")
def room_selection_get() -> dict[str, list[int]]:
    r: Redis = current_app.config["redis"]
    room = session.get("token")
    result = r.hgetall(f"room:{room}:selection")
    if "0" in result:
        return {k: json.loads(v) for k, v in result.items()}
    return {"0": []}


@io.on("room:step:set")
def room_step_set(step: int):
    print(f"setting step to {step}")
    r: Redis = current_app.config["redis"]
    room = session.get("token")
    r.set(f"room:{room}:step", step)

    emit("room:step:set", step, to=room, include_self=False)


@io.on("room:step:get")
def room_step_get() -> int:
    r: Redis = current_app.config["redis"]
    room = session.get("token")
    return r.get(f"room:{room}:step") or 0


@io.on("room:points:set")
def room_points_set(data: dict):
    print(f"setting points to {data}")
    r: Redis = current_app.config["redis"]
    room = session.get("token")
    r.hmset(f"room:{room}:points", {k: json.dumps(v) for k, v in data.items()})

    emit("room:points:set", data, to=room, include_self=False)
    # TODO: add rotation! save position and rotation and scale?


@io.on("room:points:get")
def room_points_get() -> dict[str, list[list]]:
    r: Redis = current_app.config["redis"]
    room = session.get("token")
    result: dict[int, list[list]] = r.hgetall(f"room:{room}:points")
    # TODO: type consistency!!
    result = {k: json.loads(v) for k, v in result.items()}
    if "0" not in result:
        return {"0": []}
    return result


@io.on("room:bookmarks:set")
def room_bookmarks_set(data: dict):
    r: Redis = current_app.config["redis"]
    room = session.get("token")
    r.hmset(f"room:{room}:bookmarks", data)
    emit("room:bookmarks:set", data, to=room, include_self=False)


@io.on("room:bookmarks:get")
def room_bookmarks_get() -> dict:
    r: Redis = current_app.config["redis"]
    room = session.get("token")
    return r.hgetall(f"room:{room}:bookmarks")


@io.on("room:camera:set")
def room_camera_set(data: dict):
    r: Redis = current_app.config["redis"]
    room = session.get("token")
    r.set(f"room:{room}:camera", json.dumps(data))
    emit("room:camera:set", data, to=room, include_self=False)


@io.on("room:camera:get")
def room_camera_get() -> dict:
    r: Redis = current_app.config["redis"]
    room = session.get("token")
    return json.loads(r.get(f"room:{room}:camera"))



import logging
import typing as t

import ase.io
import tqdm
import znframe
import znsocket
from celery import shared_task
from redis import Redis
from socketio import SimpleClient

from zndraw.base import FileIO

log = logging.getLogger(__name__)


@shared_task
def run_znsocket_server(port: int) -> None:
    znsocket.Server(port=port).run()
    log.critical("ZnSocket server closed.")


@shared_task
def read_file(fileio: dict, io_port: int, storage: str) -> None:
    file_io = FileIO(**fileio)
    # r = Redis(host="localhost", port=6379, db=0, decode_responses=True)

    if storage.startswith("redis"):
        r = Redis.from_url(storage, decode_responses=True)
    elif storage.startswith("znsocket"):
        r = znsocket.Client.from_url(storage)

    io = SimpleClient()

    # r = znsocket.Client("http://127.0.0.1:5000")

    # TODO: make everyone join room main
    # send update here to everyone in room, because this is only called once in the beginnig
    # chain this with compute_bonds. So this will load much faster
    r.delete("room:default:frames")

    lst = znsocket.List(r, "room:default:frames")

    if file_io.name is None:

        def _generator():
            yield ase.Atoms()

        generator = _generator()
    elif file_io.remote is not None:
        node_name, attribute = file_io.name.split(".", 1)
        try:
            import zntrack

            node = zntrack.from_rev(node_name, remote=file_io.remote, rev=file_io.rev)
            generator = getattr(node, attribute)
        except ImportError as err:
            raise ImportError(
                "You need to install ZnTrack to use the remote feature (or `pip install zndraw[all]`)."
            ) from err
    elif file_io.name.endswith((".h5", ".hdf5", ".h5md")):
        try:
            import znh5md

            reader = znh5md.ASEH5MD(file_io.name)
            generator = reader.get_atoms_list()
        except ImportError as err:
            raise ImportError(
                "You need to install ZnH5MD to use the remote feature (or `pip install zndraw[all]`)."
            ) from err
    else:
        generator = ase.io.iread(file_io.name)

    generator: t.Iterable[ase.Atoms]

    for idx, atoms in tqdm.tqdm(enumerate(generator)):
        if file_io.start and idx < file_io.start:
            continue
        if file_io.stop and idx >= file_io.stop:
            break
        if file_io.step and idx % file_io.step != 0:
            continue
        frame = znframe.Frame.from_atoms(atoms)
        lst.append(frame.to_json())
        if idx == 0:
            io.connect(f"http://127.0.0.1:{io_port}")
            io.emit("room:all:frames:refresh", [0])
        # TODO: trigger length refresh


@shared_task
def run_modifier(url, room, data: dict) -> None:
    from zndraw import ZnDraw
    from zndraw.modify import Modifier

    # cls = get_cls_from_json_schema(modifier["schema"], modifier["name"])
    vis = ZnDraw(url=url, token=room)
    vis.socket.emit("room:modifier:queue", 0)
    try:
        modifier = Modifier(**data)
        modifier.run(vis)
    except Exception as e:
        vis.log(str(e))
    finally:
        # vis[list(range(10))] = [ase.build.molecule("H2O") for _ in range(10)]
        # TODO: why is everything after 10 configuration removed?
        vis.socket.emit("room:modifier:queue", -1)

    # 1. run modifier and update redis
    # 2. use to update frames in real time "room:frames:refresh"


@shared_task
def run_selection(url, room, data: dict) -> None:
    from zndraw import ZnDraw
    from zndraw.select import Selection

    vis = ZnDraw(url=url, token=room)
    vis.socket.emit("room:selection:queue", 0)
    try:
        selection = Selection(**data)
        selection.run(vis)
    finally:
        vis.socket.emit("room:selection:queue", -1)


@shared_task
def run_analysis(url, room, data: dict) -> None:
    from zndraw import ZnDraw
    from zndraw.analyse import Analysis

    vis = ZnDraw(url=url, token=room)
    vis.socket.emit("room:analysis:queue", 0)
    try:
        analysis = Analysis(**data)
        analysis.run(vis)
    except Exception as e:
        vis.log(str(e))
    finally:
        vis.socket.emit("room:analysis:queue", -1)


@shared_task
def run_geometry(url, room, data: dict) -> None:
    from zndraw import ZnDraw
    from zndraw.draw import Geometry

    vis = ZnDraw(url=url, token=room)
    vis.socket.emit("room:geometry:queue", 0)
    try:
        geom = Geometry(**data)
        # TODO: set the position / rotation / scale
        geom.run(vis)
    finally:
        vis.socket.emit("room:geometry:queue", -1)



import io
import typing as t

import ase.io
import MDAnalysis as mda
from pydantic import BaseModel, Field


class MDAnalysis(BaseModel):
    """Select Particles using MDAnalysis selection syntax."""

    discriminator: t.Literal["MDAnalysis"] = Field("MDAnalysis")
    selection: str = Field(..., description="MDAnalysis selection string")
    append: bool = Field(False, description="Append to current selection")

    def run(self, vis) -> None:
        with io.StringIO() as f:
            ase.io.write(f, vis[vis.step], format="xyz")
            u = mda.Universe(f, format="XYZ", in_memory=True)

        selection = u.select_atoms(self.selection).ids.tolist()
        selection = [i - 1 for i in selection]
        if self.append:
            vis.selection = list(set(vis.selection + selection))
        else:
            vis.selection = selection


from zndraw.settings import _SELECTION_FUNCTIONS

_SELECTION_FUNCTIONS.append("zndraw.select.mda.MDAnalysis")



import typing as t

import numpy as np
from pydantic import BaseModel, Field
from rdkit2ase import pack, smiles2atoms

from zndraw.settings import _MODIFY_FUNCTIONS

if t.TYPE_CHECKING:
    from zndraw.zndraw import ZnDraw

_MODIFY_FUNCTIONS.append("zndraw.modify.extras.AddFromSMILES")
_MODIFY_FUNCTIONS.append("zndraw.modify.extras.Solvate")


class AddFromSMILES(BaseModel):
    """Place a molecule from a SMILES at all points."""

    discriminator: t.Literal["AddFromSMILES"] = Field("AddFromSMILES")
    SMILES: str = Field(..., description="SMILES string of the molecule to add")

    def run(self, vis: "ZnDraw", **kwargs) -> None:
        molecule = smiles2atoms(self.SMILES)

        scene = vis.atoms

        points = vis.points
        if len(points) == 0:
            points = [np.array([0, 0, 0])]

        for point in points:
            molecule_copy = molecule.copy()
            molecule_copy.translate(point)
            scene.extend(molecule_copy)

        if hasattr(scene, "connectivity"):
            del scene.connectivity

        vis.append(scene)


class Solvate(BaseModel):
    """Solvate the current scene."""

    discriminator: t.Literal["Solvate"] = Field("Solvate")
    solvent: str = Field(..., description="Solvent to use (SMILES)")
    count: int = Field(10, ge=1, le=500, description="Number of solvent molecules to add")
    density: float = Field(789, description="Density of the solvent")
    pbc: bool = Field(True, description="Whether to use periodic boundary conditions")
    tolerance: float = Field(2.0, description="Tolerance for the solvent")

    def run(self, vis: "ZnDraw", **kwargs) -> None:
        scene = vis.atoms
        scene = pack(
            [(scene, 1), (self.solvent, self.count)],
            density=self.density,
            pbc=self.pbc,
            tolerance=self.tolerance,
        )
        vis.append(scene)



import dataclasses
import os
import platform
import subprocess
import time
import uuid
import webbrowser

from celery import Celery, Task
from flask import Flask
from flask_socketio import SocketIO
from redis import Redis
from znsocket.exceptions import ConnectionError

from zndraw.settings import CeleryFileSystemConfig
from zndraw.utils import get_port

from .base import FileIO
from .tasks import read_file, run_znsocket_server

socketio = SocketIO()


def celery_init_app(app: Flask) -> Celery:
    class FlaskTask(Task):
        def __call__(self, *args: object, **kwargs: object) -> object:
            with app.app_context():
                return self.run(*args, **kwargs)

    celery_app = Celery(app.name, task_cls=FlaskTask)
    celery_app.config_from_object(app.config["CELERY"])
    celery_app.set_default()
    app.extensions["celery"] = celery_app
    return celery_app


def setup_worker() -> list:
    """Setup the worker."""
    my_env = os.environ.copy()
    if platform.system() == "Darwin" and platform.processor() == "arm":
        # fix celery worker issue on apple silicon
        my_env["OBJC_DISABLE_INITIALIZE_FORK_SAFETY"] = "YES"

    worker = subprocess.Popen(
        [
            "celery",
            "-A",
            "zndraw.make_celery",
            "worker",
            "--loglevel=info",
            # "--concurrency=6",
            # "--hostname=default_worker",
            # "--queues=celery",
            # "--prefetch-multiplier=2",
        ],
        env=my_env,
    )
    return [worker]


def create_app() -> Flask:
    """Create the Flask app."""

    app = Flask(__name__)
    from .server import main as main_blueprint

    # values to be overwritten by the server
    # these are the default
    app.config["SECRET_KEY"] = str(uuid.uuid4())

    app.config["TUTORIAL"] = ""
    app.config["AUTH_TOKEN"] = ""
    app.config["USE_TOKEN"] = True
    app.config["PORT"] = 1234
    # where is the port used?

    app.config["upgrade_insecure_requests"] = False
    app.config["compute_bonds"] = True

    app.register_blueprint(main_blueprint)

    socketio.init_app(app, cors_allowed_origins="*")

    app.config.from_mapping(
        CELERY=CeleryFileSystemConfig().to_dict(),
    )
    app.config.from_prefixed_env()
    celery_init_app(app)
    return app


@dataclasses.dataclass
class ZnDrawServer:
    use_token: bool
    upgrade_insecure_requests: bool
    compute_bonds: bool
    tutorial: str
    auth_token: str
    storage: str
    port: int = 1234
    fileio: FileIO = None
    simgen: bool = False
    celery_worker: bool = True

    _workers: list = None
    _znsocket_server: subprocess.Popen | None = None
    app: Flask = None

    def __enter__(self):
        self.app = create_app()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._workers is None:
            return
        for worker in self._workers:
            worker.kill()
        for worker in self._workers:
            worker.wait()

        # I'm not certain this will really do anything to stop the workers

        self.app.config["redis"].flushall()

    def update_cache(self):
        self.app.config["SECRET_KEY"] = str(uuid.uuid4())

        if self.storage.startswith("redis"):
            self.app.config["redis"] = Redis.from_url(self.storage, decode_responses=True)
        elif self.storage.startswith("znsocket"):
            import znsocket

            for _ in range(100):  # try to connect to znsocket for 10 s
                # if we start znsocket via celery it will take some time to start
                try:
                    self.app.config["redis"] = znsocket.Client.from_url(self.storage)
                    break
                except ConnectionError:
                    time.sleep(0.1)

        self.app.config["TUTORIAL"] = self.tutorial
        self.app.config["AUTH_TOKEN"] = self.auth_token
        self.app.config["USE_TOKEN"] = self.use_token
        self.app.config["PORT"] = self.port
        self.app.config["SIMGEN"] = self.simgen

        self.app.config["upgrade_insecure_requests"] = self.upgrade_insecure_requests
        self.app.config["compute_bonds"] = self.compute_bonds

        self.app.config["FileIO"] = self.fileio.to_dict() if self.fileio else {}

    def run(self, browser=False):
        if self.celery_worker:
            self._workers = setup_worker()

        if self.storage is None:
            port = get_port(default=3018)
            self.storage = f"znsocket://127.0.0.1:{port}"
            run_znsocket_server.delay(port)

        self.update_cache()

        read_file.delay(
            fileio=self.fileio.to_dict(), io_port=self.port, storage=self.storage
        )

        if browser:
            webbrowser.open(self.url_root)
        socketio.run(self.app, port=self.port, host="0.0.0.0")

    @property
    def url_root(self):
        return f"http://127.0.0.1:{self.port}"



import ase
import networkx as nx
import numpy as np
from ase.neighborlist import natural_cutoffs
from networkx.exception import NetworkXError
from pydantic import BaseModel, Field


class ASEComputeBonds(BaseModel):
    single_bond_multiplier: float = Field(1.2, le=2, ge=0)
    double_bond_multiplier: float = Field(0.9, le=1, ge=0)
    triple_bond_multiplier: float = Field(0.0, le=1, ge=0)

    def build_graph(self, atoms: ase.Atoms):
        cutoffs = [
            self.single_bond_multiplier,
            self.double_bond_multiplier,
            self.triple_bond_multiplier,
        ]
        atoms_copy = atoms.copy()
        connectivity_matrix = np.zeros((len(atoms_copy), len(atoms_copy)), dtype=int)
        atoms_copy.pbc = False
        distance_matrix = atoms_copy.get_all_distances(mic=False)
        np.fill_diagonal(distance_matrix, np.inf)
        for cutoff in cutoffs:
            cutoffs = np.array(natural_cutoffs(atoms_copy, mult=cutoff))
            cutoffs = cutoffs[:, None] + cutoffs[None, :]
            connectivity_matrix[distance_matrix <= cutoffs] += 1
        G = nx.from_numpy_array(connectivity_matrix)
        return G

    def update_graph_using_modifications(self, atoms: ase.Atoms):
        modifications = atoms.info.get("modifications", {})
        graph = atoms.connectivity
        for key in modifications:
            atom_1, atom_2 = key
            weight = modifications[key]
            if weight == 0:
                self.remove_edge(graph, atom_1, atom_2)
            else:
                graph.add_edge(atom_1, atom_2, weight=weight)

    @staticmethod
    def remove_edge(graph, atom_1, atom_2):
        try:
            graph.remove_edge(atom_1, atom_2)
        except NetworkXError:
            pass

    def get_bonds(self, atoms: ase.Atoms):
        graph = atoms.connectivity
        bonds = []
        for edge in graph.edges:
            bonds.append((edge[0], edge[1], graph.edges[edge]["weight"]))
        return bonds

    def update_bond_order(self, atoms: ase.Atoms, particles: list[int], order: int):
        if len(particles) != 2:
            raise ValueError("Exactly two particles must be selected")
        modifications = atoms.info.get("modifications", {})
        sorted_particles = tuple(sorted(particles))
        modifications[sorted_particles] = order
        atoms.info["modifications"] = modifications



import os
from typing import Optional

import typer

from zndraw.app import FileIO, ZnDrawServer
from zndraw.utils import get_port

cli = typer.Typer()


@cli.command()
def main(
    filename: Optional[str] = typer.Argument(
        None,
        help="Path to the file which should be visualized in ZnDraw. Can also be the name and attribute of a ZnTrack Node like 'MyNode.atoms' if at least '--remote .' is provided. ",
    ),
    webview: bool = typer.Option(
        True,
        help="""Whether to use the webview library to open the ZnDraw GUI.
        If `pip install pywebview` is available, webview will be used.
        Otherwise, the GUI will be opened in the default web browser.""",
    ),
    fullscreen: bool = typer.Option(
        False,
        help="Use fullscreen mode for the ZnDraw GUI. (only with webview)",
    ),
    port: int = typer.Option(
        None, help="""Port to use for the ZnDraw server. Default port is 1234"""
    ),
    browser: bool = typer.Option(
        True, help="""Whether to open the ZnDraw GUI in the default web browser."""
    ),
    start: int = typer.Option(
        0,
        help="""First frame to be visualized. If set to 0, the first frame will be visualized.""",
    ),
    stop: int = typer.Option(
        None,
        help="""Last frame to be visualized. If set to None, the last frame will be visualized.""",
    ),
    step: int = typer.Option(
        1,
        help="""Stepsize for the frames to be visualized. If set to 1, all frames will be visualized.
        If e.g. set to 2, every second frame will be visualized.""",
    ),
    compute_bonds: bool = typer.Option(
        True,
        help="""Whether to compute bonds for the structure. If set to False, no bonds will be computed.""",
    ),
    upgrade_insecure_requests: bool = typer.Option(
        False,
        hidden=True,
        help="Set the html attribute upgrade-insecure-requests. If you are running ZnDraw behind a reverse proxy and encounter issues with insecure requests, you might want to set this to true.",
    ),
    use_token: bool = typer.Option(
        True,
        help="DEPRECATED!! Use a token to authenticate the ZnDraw server.  This is useful if you are running ZnDraw as a server application.",
    ),
    remote: str = typer.Option(
        None,
        help="URL to a ZnTrack repository to stream data from.",
    ),
    rev: str = typer.Option(
        None,
        help="Revision of the ZnTrack repository to stream data from.",
    ),
    tutorial: str = typer.Option(
        None,
        help="Show the tutorial from the URL inside an IFrame.",
    ),
    auth_token: str = typer.Option(
        None,
        help="Token to authenticate pyclient requests to the ZnDraw server, e.g., for adding defaults to all webclients.",
    ),
    simgen: bool = typer.Option(
        False,
        help="Show the SiMGen demo UI.",
    ),
    celery_worker: bool = typer.Option(
        True,
        help="Start ZnDraw with celery workers. If disabled, you must manage the workers yourself. This can be useful when hosting ZnDraw for multiple users.",
    ),
    storage: str = typer.Option(
        None,
        help="URL to the redis `redis://localhost:6379/0` or znsocket `znsocket://127.0.0.1:6379` server. If None is provided, a local znsocket server will be started.",
    ),
):
    """Start the ZnDraw server.

    Visualize Trajectories, Structures, and more in ZnDraw.
    """
    if port is None:
        port = get_port()

    fileio = FileIO(
        name=filename,
        remote=remote,
        rev=rev,
        start=start,
        stop=stop,
        step=step,
    )
    if "ZNDRAW_STORAGE" in os.environ and storage is None:
        storage = os.environ["ZNDRAW_STORAGE"]

    with ZnDrawServer(
        use_token=use_token,
        upgrade_insecure_requests=upgrade_insecure_requests,
        compute_bonds=compute_bonds,
        tutorial=tutorial,
        auth_token=auth_token,
        port=port,
        fileio=fileio,
        simgen=simgen,
        celery_worker=celery_worker,
        storage=storage,
    ) as app:
        app.run(browser=browser)



import logging
import typing as t

import pydantic

from zndraw.utils import ensure_path

log = logging.getLogger(__name__)


class CeleryBaseConfig(pydantic.BaseModel):
    name: t.Literal["CeleryBaseConfig"] = "CeleryBaseConfig"
    broker: str = "filesystem://"
    result_backend: str = "cache"
    cache_backend: str = "memory"
    task_ignore_result: bool = True

    @property
    def task_routes(self):
        return {
            "*._run_global_modifier": {"queue": "slow"},
            "*.update_atoms": {"queue": "fast"},
            "*.get_selection_schema": {"queue": "fast"},
            "*.scene_schema": {"queue": "fast"},
            "*.geometries_schema": {"queue": "fast"},
            "*.analysis_schema": {"queue": "fast"},
            "*.handle_room_get": {"queue": "fast"},
            "*.activate_modifier": {"queue": "fast"},
            "*on_disconnect": {"queue": "fast"},
            "*upload_file": {"queue": "fast"},
        }

    def to_dict(self):
        return dict(
            broker_url=self.broker,
            result_backend=self.result_backend,
            cache_backend=self.cache_backend,
            task_ignore_result=self.task_ignore_result,
            task_routes=self.task_routes,
        )


class CeleryFileSystemConfig(CeleryBaseConfig):
    name: t.Literal["CeleryFileSystemConfig"] = "CeleryFileSystemConfig"
    data_folder: str = "~/.zincware/zndraw/celery/out"
    data_folder_processed: str = "~/.zincware/zndraw/celery/processed"
    task_ignore_result: bool = True

    def to_dict(self):
        return dict(
            broker_url=self.broker,
            broker_transport_options=dict(
                data_folder_in=ensure_path(self.data_folder),
                data_folder_out=ensure_path(self.data_folder),
                data_folder_processed=ensure_path(self.data_folder_processed),
            ),
            result_backend=self.result_backend,
            cache_backend=self.cache_backend,
            task_ignore_result=self.task_ignore_result,
            task_routes=self.task_routes,
        )



import logging
import uuid

from flask import (
    current_app,
    redirect,
    request,
    send_from_directory,
    session,
)

from . import main

log = logging.getLogger(__name__)


def _upload(file, url, token):
    import pathlib

    import ase.io

    from zndraw.zndraw import ZnDraw

    vis = ZnDraw(url=url, token=token)
    for atoms in ase.io.iread(pathlib.Path("data", file)):
        vis.append(atoms)


@main.route("/")
def index():
    """Render the main ZnDraw page."""
    try:
        token = session["token"]
    except KeyError:
        token = uuid.uuid4().hex[:8] if current_app.config["USE_TOKEN"] else "main"
        session["token"] = token

    session["name"] = uuid.uuid4().hex[:8]
    # just show templaces / index.html
    return send_from_directory("templates", "index.html")


@main.route("/<path:filename>")
def main_files(filename):
    return send_from_directory("templates/", filename)


@main.route("/assets/<path:filename>")
def assets(filename):
    return send_from_directory("templates/assets", filename)

    # return render_template(
    #     "index.jinja2",
    #     upgrade_insecure_requests=current_app.config["upgrade_insecure_requests"],
    #     token=session["token"],
    #     tutorial=current_app.config["TUTORIAL"],
    #     simgen=current_app.config["SIMGEN"],
    # )


@main.route("/token/<token>")
def token(token):
    session["token"] = token
    return redirect("/")


@main.route("/reset")
def reset():
    session["token"] = uuid.uuid4().hex[:8]  # TODO: how should reset work locally?
    return redirect("/")


@main.route("/exit")
@main.route("/exit/<token>")
def exit_route(token: str = None):
    """Exit the session."""
    log.critical("Server shutting down...")
    if token != current_app.config["AUTH_TOKEN"]:
        return "Invalid auth token", 403

    from ..app import socketio

    socketio.stop()
    return "Server shutting down..."


@main.route("/file/<file>")
def file(file: str):
    """Open a file on the server."""
    # open a new connection, read the file, upload it to the server, and then close the connection
    import multiprocessing as mp

    try:
        token = session["token"]
    except KeyError:
        token = uuid.uuid4().hex[:8] if current_app.config["USE_TOKEN"] else "main"
        session["token"] = token
    url = request.url_root
    print(f"URL: {url}")

    proc = mp.Process(target=_upload, args=(file, url, token), daemon=True)
    proc.start()
    # TODO: why is this not using celery?

    return redirect(url)



import enum
import io
import typing as t

import ase.io
import MDAnalysis as mda
import numpy as np
import plotly.express as px
from MDAnalysis.analysis import msd, rdf
from pydantic import BaseModel, Field

if t.TYPE_CHECKING:
    from zndraw import ZnDraw


class MDAInterRDFNorm(str, enum.Enum):
    """Normalization of RDF."""

    rdf = "rdf"
    density = "density"
    none = "none"


class MDAInterRDF(BaseModel):
    """Select Particles using MDAnalysis selection syntax."""

    discriminator: t.Literal["MDAInterRDF"] = Field("MDAInterRDF")
    selection_a: str = Field(..., description="MDAnalysis selection string")
    selection_b: str = Field(..., description="MDAnalysis selection string")
    nbins: int = Field(100, description="Number of bins")
    range: str = Field("0, 10", description="Range of histogram")
    norm: MDAInterRDFNorm = Field(MDAInterRDFNorm.rdf, description="Normalization of RDF")

    def run(self, vis) -> None:
        self.range = tuple(map(float, self.range.split(",")))
        with io.StringIO() as f:
            ase.io.write(f, list(vis), format="xyz")
            u = mda.Universe(f, format="XYZ", in_memory=True)

        grp1 = u.select_atoms(self.selection_a)
        grp2 = u.select_atoms(self.selection_b)

        box = np.diag(vis[vis.step].cell)
        u.dimensions = [box[0], box[1], box[2], 90, 90, 90]

        result = (
            rdf.InterRDF(
                grp1, grp2, nbins=self.nbins, range=self.range, norm=self.norm.value
            )
            .run()
            .results
        )

        fig = px.line(
            x=result.bins,
            y=result.rdf,
            # title="Distance between selected particles",
            render_mode="svg",  # This is important, otherwise openGL will be used
            # and there can/will be issues with three.js
        )

        vis.figure = fig.to_json()


class MDAEinsteinMSD(BaseModel):
    discriminator: t.Literal["MDAEinsteinMSD"] = Field("MDAEinsteinMSD")
    selection: str = Field("all", description="MDAnalysis selection string")
    delta_t: float = Field(0.1, description="Time step")

    def run(self, vis: "ZnDraw") -> None:
        vis.log("Running EinsteinMSD - Loading trajectory...")
        with io.StringIO() as f:
            ase.io.write(f, list(vis), format="xyz")
            u = mda.Universe(f, format="XYZ", in_memory=True)
        box = np.diag(vis[vis.step].cell)
        u.dimensions = [box[0], box[1], box[2], 90, 90, 90]
        # u.trajectory.dt = self.delta_t

        vis.log("Running EinsteinMSD - Running analysis...")
        MSD = msd.EinsteinMSD(u, select=self.selection, msd_type="xyz", fft=True)
        MSD.run()

        nframes = MSD.n_frames
        lagtimes = np.arange(nframes) * self.delta_t

        fig = px.line(
            x=lagtimes,
            y=MSD.results.timeseries,
            render_mode="svg",  # This is important, otherwise openGL will be used
            # and there can/will be issues with three.js
        )
        vis.figure = fig.to_json()


from zndraw.settings import _ANALYSIS_FUNCTIONS

_ANALYSIS_FUNCTIONS.append("zndraw.analyse.mda.MDAInterRDF")
_ANALYSIS_FUNCTIONS.append("zndraw.analyse.mda.MDAEinsteinMSD")



import enum

from pydantic import BaseModel, Field


class Material(str, enum.Enum):
    MeshBasicMaterial = "MeshBasicMaterial"
    MeshLambertMaterial = "MeshLambertMaterial"
    MeshMatcapMaterial = "MeshMatcapMaterial"
    MeshPhongMaterial = "MeshPhongMaterial"
    MeshPhysicalMaterial = "MeshPhysicalMaterial"
    MeshStandardMaterial = "MeshStandardMaterial"
    MeshToonMaterial = "MeshToonMaterial"


# create a class for the material, resolution, etc.
class Scene(BaseModel):
    fps: int = Field(30, ge=1, le=120, description="Maxium frames per second")
    # material: Material = Field(Material.MeshPhongMaterial, description="Material")
    # resolution: int = Field(10, ge=1, le=50, description="Resolution")
    # particle_size: float = Field(1.0, ge=0.1, le=5, description="Particle Size")
    # bonds_size: float = Field(1.0, ge=0.1, le=5, description="Bonds Size")
    # wireframe: bool = Field(False, description="Wireframe")
    loop: bool = Field(
        False,
        alias="Animation Loop",
        description="Automatically restart animation when finished.",
    )
    simulation_box: bool = Field(
        False,
        description="Show the simulation box.",
    )
    # bonds: bool = Field(
    #     True,
    #     description="Show bonds.",
    # )
    # line_label: bool = Field(
    #     True,
    #     description="Show the length of the line.",
    # )
    # label_offset: int = Field(
    #     0,
    #     ge=-7,
    #     le=7,
    #     description="Move the label to the left or right (keypress i).",
    # )

    @classmethod
    def updated_schema(cls) -> dict:
        schema = cls.model_json_schema()

        # schema["properties"]["wireframe"]["format"] = "checkbox"
        schema["properties"]["Animation Loop"]["format"] = "checkbox"
        schema["properties"]["simulation_box"]["format"] = "checkbox"
        # schema["properties"]["resolution"]["format"] = "range"
        # schema["properties"]["label_offset"]["format"] = "range"
        # schema["properties"]["particle_size"]["format"] = "range"
        schema["properties"]["fps"]["format"] = "range"
        # schema["properties"]["particle_size"]["step"] = 0.1
        # schema["properties"]["bonds_size"]["format"] = "range"
        # schema["properties"]["bonds_size"]["step"] = 0.1
        # schema["properties"]["bonds"]["format"] = "checkbox"
        # schema["properties"]["line_label"]["format"] = "checkbox"

        return schema



import contextlib
import importlib.util
import json
import logging
import pathlib
import socket
import sys
import tempfile
import uuid
from typing import get_type_hints

import ase
import datamodel_code_generator
import numpy as np
import znframe
from decorator import decorator

SHARED = {"atoms": None}


def split_list_into_chunks(lst, n):
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i : i + n]


def estimate_max_batch_size_for_socket(frames: list[znframe.Frame]):
    from .settings import GlobalConfig

    max_size = GlobalConfig().max_socket_data_size
    sizes = [
        sys.getsizeof(json.dumps(frame.to_dict(built_in_types=False))) for frame in frames
    ]
    largest_frame = max(sizes)
    chunk_size = int(max_size / largest_frame * 0.9)
    return max(chunk_size, 1)


def rgb2hex(value):
    r, g, b = np.array(value * 255, dtype=int)
    return "#%02x%02x%02x" % (r, g, b)


def get_radius(value):
    return (0.25 * (2 - np.exp(-0.2 * value)),)


def get_port(default: int = 1234) -> int:
    """Get an open port."""
    try:
        sock = socket.socket()
        sock.bind(("", default))
        port = default
    except OSError:
        sock = socket.socket()
        sock.bind(("", 0))
        port = sock.getsockname()[1]
    finally:
        sock.close()
    return port


@contextlib.contextmanager
def set_global_atoms(atoms: ase.Atoms):
    """Temporarily create a 'SHARED["atoms"]' variable."""
    # TODO: I know this is bad, but I don't know how to do it better - send help!
    SHARED["atoms"] = atoms
    yield
    SHARED["atoms"] = None


class ZnDrawLoggingHandler(logging.Handler):
    """Logging handler which emits log messages to the ZnDraw server."""

    def __init__(self, zndraw):
        super().__init__()
        self.zndraw = zndraw

    def emit(self, record):
        try:
            msg = self.format(record)
            self.zndraw.log(msg)
        except RecursionError:  # See StreamHandler
            raise
        except Exception:
            print("Something went wrong")
            self.handleError(record)


def get_cls_from_json_schema(schema: dict, name: str, **kwargs):
    """Get a python class from a json schema."""

    # TODO: needs tests
    # TODO: do not write file but use in-memory

    kwargs["strict_nullable"] = True

    with tempfile.TemporaryDirectory() as temporary_directory_name:
        temporary_directory = pathlib.Path(temporary_directory_name)
        output = temporary_directory / "model.py"
        datamodel_code_generator.generate(
            json.dumps(schema),
            input_file_type=datamodel_code_generator.InputFileType.JsonSchema,
            input_filename="example.json",
            output=output,
            # set up the output model types
            output_model_type=datamodel_code_generator.DataModelType.PydanticV2BaseModel,
            **kwargs,
        )

        ref_module = uuid.uuid4().hex

        spec = importlib.util.spec_from_file_location(ref_module, output)
        module = importlib.util.module_from_spec(spec)
        sys.modules[ref_module] = module
        spec.loader.exec_module(module)

        return getattr(module, name)


def hide_discriminator_field(d: dict) -> None:
    """Inplace method to set discriminator field to hidden."""
    raise ValueError("This method is not used anymore!")
    for v in d.values():
        if isinstance(v, dict):
            if v.get("title") == "Discriminator":
                v["options"] = {"hidden": True}
                v["type"] = "string"
            hide_discriminator_field(v)


@decorator
def typecast(func, *args, **kwargs):
    annotations = get_type_hints(func)
    updated_args = []
    updated_kwargs = {}
    for arg, arg_type in zip(args, annotations.values()):
        if not isinstance(arg, arg_type):
            if isinstance(arg, dict):
                updated_args.append(arg_type(**arg))
            else:
                updated_args.append(arg_type(arg))
        else:
            updated_args.append(arg)

    for kwarg, kwarg_type in annotations.items():
        if kwarg in kwargs:
            if not isinstance(kwargs[kwarg], kwarg_type):
                if isinstance(kwargs[kwarg], dict):
                    updated_kwargs[kwarg] = kwarg_type(**kwargs[kwarg])
                else:
                    updated_kwargs[kwarg] = kwarg_type(kwargs[kwarg])
            else:
                updated_kwargs[kwarg] = kwargs[kwarg]
    return func(*updated_args, **updated_kwargs)


@decorator
def typecast_kwargs(func, **kwargs):
    annotations = get_type_hints(func)
    arg_type = list(annotations.values())[0]
    return arg_type(**kwargs)


def ensure_path(path: str):
    """Ensure that a path exists."""
    p = pathlib.Path(path).expanduser()
    p.mkdir(parents=True, exist_ok=True)
    return p.as_posix()


def wrap_and_check_index(index: int | slice | list[int], length: int) -> list[int]:
    is_slice = isinstance(index, slice)
    if is_slice:
        index = list(range(*index.indices(length)))
    index = [index] if isinstance(index, int) else index
    index = [i if i >= 0 else length + i for i in index]
    # check if index is out of range
    for i in index:
        if i >= length:
            raise IndexError(f"Index {i} out of range for length {length}")
        if i < 0:
            raise IndexError(f"Index {i-length} out of range for length {length}")
    return index


def check_selection(value: list[int], maximum: int):
    """Check if the selection is valid

    Attributes
    ----------
        value: list[int]
            the selected indices
        maximum: int
            len(vis.step), will be incremented by one, to account for
    """
    if not isinstance(value, list):
        raise ValueError("Selection must be a list")
    if any(not isinstance(i, int) for i in value):
        raise ValueError("Selection must be a list of integers")
    if len(value) != len(set(value)):
        raise ValueError("Selection must be unique")
    if any(i < 0 for i in value):
        raise ValueError("Selection must be positive integers")
    if any(i >= maximum for i in value):
        raise ValueError(
            f"Can not select particles indices larger than size of the scene: {maximum }. Got {value}"
        )
